/*
 *  GameApp.cpp
 *  
 *
 *  Created by Mark Szymczyk
 *  Copyright (c) 2006 Me and Mark Publishing. All rights reserved.
 *
 */

#include "GameApp.h"

// Constructor
GameApp::GameApp(void)
{

}

// Destructor
GameApp::~GameApp(void)
{
    CleanUpApp();
}

// Accessors
GameContextPtr GameApp::GetContext(void)
{
    return context;
}

void GameApp::SetContext(GameContextPtr theContext)
{
    context = theContext;
}

GameLevelPtr GameApp::GetCurrentLevel(void)
{
    return currentLevel;
}

void GameApp::SetCurrentLevel(GameLevelPtr theLevel)
{
    currentLevel = theLevel;
}

void GameApp::InitApp(void)
{
    InitToolbox();
    SetUpMenuBar();
    InstallKeyboardEventHandler();
    InstallTimer();

    // Fit 32 tiles horizontally and 24 tiles vertically on the screen.
    InitializeDrawContext(32, 24);
    
    LoadLevel(CFSTR("Level1"), CFSTR("rsrc"), NULL, kLevelResourceID);
    LoadTileList(CFSTR("TileList"), CFSTR("rsrc"), NULL, kTileResourceID);
        
   }

void GameApp::InitToolbox (void)
{
    // Initialize all the needed managers. 
    // In Carbon, you need to initialize only the cursor.
	
    //InitCursor();
		
}

void GameApp::InstallKeyboardEventHandler(void)
{
    EventTypeSpec keyboardHandlerEvents = { kEventClassKeyboard, kEventRawKeyDown };
    
    InstallApplicationEventHandler(NewEventHandlerUPP(KeyboardEventHandler), kKeyboardEventTotal,
                                   &keyboardHandlerEvents, this, NULL);
    
}

void GameApp::InstallTimer(void)
{
    EventLoopRef mainLoop;
    
    mainLoop = GetMainEventLoop();
    
    gameTimer = NewEventLoopTimerUPP(GameLoopTimer);
    
    // Have game loop run 60 times per second
    // We won't have the timer start until we go fullscreen.
    InstallEventLoopTimer(mainLoop, (kEventDurationForever), (kEventDurationSecond / 60),
            gameTimer, this, &timer);
    
}

void GameApp::LoadLevel(CFStringRef filename, CFStringRef extension, CFStringRef subdirectory, short levelID)
{
    currentLevel = new GameLevel;
	if (currentLevel == NULL)
		ExitToShell();
    
	Boolean success;
	success = currentLevel->Load(filename, extension, subdirectory, levelID);
	if (success == false)
		ExitToShell();
    
}

void GameApp::LoadTileList(CFStringRef filename, CFStringRef extension, CFStringRef subdirectory, short tileID)
{
    currentTileList = new GameTileList;
	if (currentLevel == NULL)
		ExitToShell();
    
    Boolean tilesLoaded;
    tilesLoaded = currentTileList->Load(filename, extension, subdirectory, tileID);
    if (!tilesLoaded)
        ExitToShell();
    
}

void GameApp::InitializeDrawContext(short width, short height)
{
    // Create the draw context
	context = new GameContext;
	if (context == NULL)
		ExitToShell();
  
    context->InitializeAGL();
    
    // Initialize the offsets for scrolling.
    context->SetXOffset(0);
    context->SetYOffset(0);
    
    // Create the tile storage texture map
    GameTexturePtr textureMap = new GameTexture;
        
    textureMap->SetInternalFormat(GL_RGBA8);
    
    // The call to QTNewGWorldFromPtr() requires a
    // 32-bit ARGB pixel format for 32-bit color. There
    // is no GL_ARGB format so we must use GL_BGRA and
    // reverse the values on big-endian systems.
    textureMap->SetFormat(GL_BGRA);
    
    #if __BIG_ENDIAN__
        textureMap->SetType(GL_UNSIGNED_INT_8_8_8_8_REV);
    #else
        textureMap->SetType(GL_UNSIGNED_INT_8_8_8_8);
    #endif
    
    textureMap->Load(CFSTR("Tiles"), CFSTR("png"), NULL);
    context->SetTileStorage(textureMap);
    
    // Set the size of the context
    context->SetWidthInTiles(width);
    context->SetHeightInTiles(height);
    gluOrtho2D(0.0, width, 0.0, height);
}


void GameApp::CleanUpApp(void)
{
	// Release the memory of the variables we allocated
	if (context != NULL) {
        delete context;
        context = NULL;
    }
    
	if (currentLevel != NULL) {
        delete currentLevel;
        currentLevel =  NULL;
    }
    
	if (currentTileList != NULL) {
        delete currentTileList;
        currentTileList = NULL;
    }
    
    QuitApplicationEventLoop();
    RemoveEventLoopTimer(timer);
}


// Event related functions

void GameApp::EventLoop(void)
{
    RunApplicationEventLoop();    
}

// menu related functions

void GameApp::SetUpMenuBar()
{
    OSStatus error;
    IBNibRef menuNib;
     
    // Read menu from nib file main.nib
     
    error = CreateNibReference(CFSTR("main"), &menuNib);
    if (!error)
        SetMenuBarFromNib(menuNib, CFSTR("MenuBar"));
                          
    RegisterMenuCommands();
}

void GameApp::RegisterMenuCommands(void)
{
    // Register our Start/Resume menu item    
    SetMenuItemCommandID(GetMenuRef(kFileMenu), kResumeMenuItem, kCommandResumeGame);
    
    EventTypeSpec menuHandlerEvents = 
        { kEventClassCommand, kEventCommandProcess };
    
    InstallApplicationEventHandler(NewEventHandlerUPP(MenuEventHandler), kEventTotal,
            &menuHandlerEvents, this, NULL);
}

pascal OSStatus GameApp::MenuEventHandler(EventHandlerCallRef myHandlerChain,
        EventRef event, void* userData)
{
    HICommand menuItemChosen;
    OSStatus result;

    GameAppPtr currentApp = (GameAppPtr)userData;
    
    GetEventParameter(event, kEventParamDirectObject, typeHICommand, NULL, 
            sizeof(HICommand), NULL, &menuItemChosen);
            
    // The commandID field of the HICommand data structure tells us
    // the menu command of the menu item the player chose.
    switch (menuItemChosen.commandID) {
        case kCommandResumeGame:
            currentApp->Resume();
            result = noErr;
            break;
            
        case kHICommandQuit:
            currentApp->Quit();
            result = eventNotHandledErr;
            break;
            
        default:
            // The standard application event handler
            // handles any events our menu event handler doesn't handle.
            result = eventNotHandledErr;
            break;
    }
    
    return result;
}

void GameApp::Quit(void)
{
    QuitApplicationEventLoop();	
}


// Game related functions
pascal void GameApp::GameLoopTimer(EventLoopTimerRef theTimer, void* userData)
{
    GameAppPtr currentApp = (GameAppPtr)userData;
    
    currentApp->GameLoop();
}

pascal OSStatus GameApp::KeyboardEventHandler(EventHandlerCallRef myHandlerChain,
                                          EventRef event, void* userData)
{
    UInt32 keyPressed;
    OSStatus result;
    
    GameAppPtr currentApp = (GameAppPtr)userData;
    GameContextPtr currentContext = currentApp->GetContext();
    GameLevelPtr level = currentApp->GetCurrentLevel();
    
    GetEventParameter(event, kEventParamKeyCode, typeUInt32, NULL, 
                      sizeof(UInt32), NULL, &keyPressed);
    
    switch (keyPressed) {
        case kEscapeKey:
            currentApp->Pause();
            result = noErr;
            break;

        case kLeftArrow:
            currentContext->ScrollLeft(level, kOneTile);
            result = noErr;
            break;
            
        case kRightArrow:
            currentContext->ScrollRight(level, kOneTile);
            result = noErr;
            break;

        case kUpArrow:
            currentContext->ScrollUp(level, kOneTile);
            result = noErr;
            break;

        case kDownArrow:
            currentContext->ScrollDown(level, kOneTile);
            result = noErr;
            break;
            
        default:
            // The standard application event handler
            // handles any events our menu event handler doesn't handle.
            result = eventNotHandledErr;
            break;
    }
    
    return result;
}

void GameApp::GameLoop(void)
{
    RenderFrame();
}

void GameApp::RenderFrame(void) 
{
    glClear(GL_COLOR_BUFFER_BIT);
    context->DrawTiles(currentLevel);
    aglSwapBuffers(context->GetAGLDrawContext());
}

void GameApp::Pause(void)
{
    // Turn off fullscreen
    context->TurnOffFullscreen();
    
    DrawMenuBar();
    
    // Shut off the game loop timer
    OSStatus error;
    error = SetEventLoopTimerNextFireTime(timer, (kEventDurationForever));

}

void GameApp::Resume(void)
{
    context->TurnOnFullscreen();

    // Turn on the game loop timer
    OSStatus error2;
    error2 = SetEventLoopTimerNextFireTime(timer, kEventDurationNoWait);

}
