/*
 *  GameContext.cpp
 *  MGPCode(OpenGL)
 *
 *  Created by Mark Szymczyk on 4/18/06.
 *  Copyright 2006 Me and Mark Publishing. All rights reserved.
 *
 */

#include "GameContext.h"

// Constructor and destructor
GameContext::GameContext(void)
{
    
}

GameContext::~GameContext(void)
{
    delete tileStorage;
    aglDestroyContext(aglDrawContext);
}


// Accessors
AGLContext GameContext::GetAGLDrawContext(void)
{
    return aglDrawContext;
}

void GameContext::SetAGLDrawContext(AGLContext theContext)
{
    aglDrawContext = theContext;
}

GameTexturePtr GameContext::GetTileStorage(void)
{
    return tileStorage;
}

void GameContext::SetTileStorage(GameTexturePtr theStorage)
{
    tileStorage = theStorage;
}

short GameContext::GetXOffset(void)
{
    return xOffset;
}

void GameContext::SetXOffset(short offset)
{
    xOffset = offset;
}

short GameContext::GetYOffset(void)
{
    return yOffset;
}

void GameContext::SetYOffset(short offset)
{
    yOffset = offset;
}

short GameContext::GetWidthInTiles(void)
{
    return widthInTiles;
}

void GameContext::SetWidthInTiles(short width)
{
    widthInTiles = width;
}

short GameContext::GetHeightInTiles(void)
{
    return heightInTiles;
}

void GameContext::SetHeightInTiles(short height)
{
    heightInTiles = height;
}


// AGL functions
void GameContext::InitializeAGL(void)
{
    GLint attributeList[] = { AGL_RGBA, AGL_ACCELERATED, AGL_DOUBLEBUFFER, 
        AGL_FULLSCREEN, AGL_NONE };
	AGLPixelFormat 	pixelFormat;
	
    // I had several people report problems with aglChoosePixelFormat() failing
    // on Mac OS X 10.4.7. The problem is that the original code passed NULL as the first
    // parameter and 0 as the second parameter. The AGL documentation says passing NULL as
    // the first parameter is the way to go to support multiple monitor systems, but it doesn't
    // work with the AGL_FULLSCREEN attribute. The fix is to call GetMainDevice() and pass the
    // handle GetMainDevice() returns as the first argument to aglChoosePixelFormat(). I commented
    // out the original line of code in case you need to use it.
    
    // Choose pixel format
    GLenum error;
    
    GDHandle mainScreen = GetMainDevice();
    pixelFormat = aglChoosePixelFormat(&mainScreen, 1, attributeList);
    
    //pixelFormat = aglChoosePixelFormat(NULL, 0, attributeList); 
    error = aglGetError();   
    if (pixelFormat == NULL)
        ExitToShell();
    
    aglDrawContext = aglCreateContext (pixelFormat, NULL);
    error = aglGetError();   
    if (aglDrawContext == NULL)
        ExitToShell();
    
    GLboolean success;
    success = aglSetCurrentContext(aglDrawContext);
    error = aglGetError();   
    
    aglDestroyPixelFormat(pixelFormat);
    error = aglGetError();   
    
}

void GameContext::TurnOnFullscreen(void)
{
    GLenum error;
    
    aglSetCurrentContext(aglDrawContext);
    error = aglGetError();   
    
    // Go fullscreen 1024 by 768.
    // 60 is the refresh frequency.
    // 0 is the graphics device index.  This 
    // will work for single monitor systems.
    GLboolean success;
    success = aglSetFullScreen(aglDrawContext, 1024, 768, 60, 0);
    
}

void GameContext::TurnOffFullscreen(void)
{
    aglSetDrawable(aglDrawContext, NULL);    
}


// Scrolling functions
short GameContext::ScrollLeft(GameLevelPtr theLevel, short tilesToScroll)
{
    // The scrolling functions return the number 
    // of tiles scrolled
    
	// Bounds checking
    short leftEdge = 0;
	if (xOffset <= leftEdge)
		return 0;
    
	// If a full scroll puts us past the left edge
	// then only scroll to the edge.
	if ((xOffset - tilesToScroll) < leftEdge) {
		short amountToScroll = xOffset;
        return (ScrollLeft(theLevel, amountToScroll));
	}	
 
    // Update the origin to reflect the scroll
	xOffset = xOffset - tilesToScroll;
    
    return tilesToScroll;
}

short GameContext::ScrollRight(GameLevelPtr theLevel, short tilesToScroll)
{
    // The scrolling functions return the number 
    // of tiles scrolled
    
	// Bounds checking.  If we've hit the right edge of the level, don't
	// bother scrolling.
	short rightEdge = theLevel->GetLevelWidth() - widthInTiles;
	if (xOffset >= rightEdge)
		return 0;
    
	// If a full scroll puts us past the right edge
	// then only scroll to the edge.
	if ((xOffset + tilesToScroll) > rightEdge) {
		short amountToScroll = rightEdge - xOffset;
		return (ScrollRight(theLevel, amountToScroll));
	}
    
    // Update the origin to reflect the scroll
	xOffset = xOffset + tilesToScroll;
    
    return tilesToScroll;
}

short GameContext::ScrollUp(GameLevelPtr theLevel, short tilesToScroll)
{
    // The scrolling functions return the number 
    // of tiles scrolled
    
	// Bounds checking
    short topEdge = 0;
	if (yOffset <= topEdge)
		return 0;
    
	// If a full scroll puts us past the top edge
	// then only scroll to the edge.
	if ((yOffset - tilesToScroll) < topEdge) {
		short amountToScroll = yOffset;
        return (ScrollUp(theLevel, amountToScroll));
	}
    
    // Update the origin to reflect the scroll
	yOffset = yOffset - tilesToScroll;
    
    return tilesToScroll;
}

short GameContext::ScrollDown(GameLevelPtr theLevel, short tilesToScroll)
{
    // The scrolling functions return the number 
    // of tiles scrolled
    
	// Bounds checking.  If we've hit the bottom of the level,
	// don't bother scrolling
	short bottomEdge = theLevel->GetLevelHeight() - heightInTiles;
	if (yOffset >= bottomEdge)
		return 0;
    
	// If a full scroll puts us past the bottom edge
	// then only scroll to the edge.
	if ((yOffset + tilesToScroll) > bottomEdge) {
		short amountToScroll = bottomEdge - yOffset;
        return (ScrollDown(theLevel, amountToScroll));    
	}
    
    // Update the origin to reflect the scroll
	yOffset = yOffset + tilesToScroll;
    
    return tilesToScroll;
}

// Drawing functions
void GameContext::DrawTiles(GameLevelPtr theLevel)
{
	for (short row = 0; row < heightInTiles; row++) {
		for (short column = 0; column < widthInTiles; column++) {
            DrawTileFromLevelMap(theLevel, row, column);
		}
	}
    
}

void GameContext::DrawTileFromLevelMap(GameLevelPtr theLevel, short theRow, short theColumn)
{
    // Find where we are in the level.
    short worldRow = theRow + yOffset;
    short worldColumn = theColumn + xOffset;
    
    // Determine the tile that needs to be drawn.
    short theLevelWidth = theLevel->GetLevelWidth();
	UInt32 mapIndex = (worldRow * theLevelWidth) + worldColumn;	
	short tileNumber = theLevel->levelMap[mapIndex].GetValue();
    
    // Find the tile in the tile storage.
    Point tileGridLocation = GetGridLocation(tileNumber);
    float tileGridX = tileGridLocation.h;
    float tileGridY = tileGridLocation.v;
    
    GLfloat textureLeft = 1.0f * (tileGridX / kTilesPerTextureRow);
    GLfloat textureRight = textureLeft + (1.0f / kTilesPerTextureRow);
    GLfloat textureTop = 1.0f * (tileGridY / kTilesPerTextureColumn); //((heightInTiles - tileGridLocation.v) / heightInTiles);
    GLfloat textureBottom = textureTop + (1.0f / kTilesPerTextureColumn); //- (1.0f / heightInTiles);
        
    // Determine where to plot the tile on the screen.
    // Our levels start at the top, but the OpenGL screen
    // starts at the bottom so we have to reverse the y coordinates.
    GLfloat left = theColumn;
    GLfloat right = theColumn + 1.0f;
    GLfloat top = heightInTiles - theRow;
    GLfloat bottom = top - 1.0f;
    
    // Set up the drawing of the tile
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, tileStorage->GetTextureName());
    
    
    // Draw the tile
    glBegin(GL_QUADS);
    
    // Lower left corner
    glTexCoord2f(textureLeft, textureBottom);
    glVertex3f(left, bottom, 0.0);
    
    // Lower right corner
    glTexCoord2f(textureRight, textureBottom);
    glVertex3f(right, bottom, 0.0);
    
    // Upper right corner
    glTexCoord2f(textureRight, textureTop);
    glVertex3f(right, top, 0.0);
    
    // Upper left corner
    glTexCoord2f(textureLeft, textureTop);
    glVertex3f(left, top, 0.0);
    
    glEnd();
}

// Utility functions
Point GameContext::GetGridLocation(short tileNumber)
{
	// Given a tile number from the level map, return
	// the grid location of the tile on the tile board.
	Point result;
	
	result.h = tileNumber % kTilesInRow;
	result.v = tileNumber / kTilesInRow;
	
	return result;
}
