// Level.cpp  	
// Written by Mark Szymczyk

#ifndef GAME_LEVEL
	#include "GameLevel.h"
#endif

// Map Element Accessor Functions
short MapElement::GetValue(void)
{
	return value;
}

void MapElement::SetValue(short theValue)
{
	value = theValue;
}

Boolean MapElement::IsVisibleToPlayer(void)
{
	return visibleToPlayer;
}

void MapElement::SetVisibleToPlayer(Boolean visibility)
{
	visibleToPlayer = visibility;
}


// GameLevel default constructor
GameLevel::GameLevel(void)
{
	levelWidth = 0;
	levelHeight = 0;
}

// Copy constructor
GameLevel::GameLevel(const GameLevel& theLevel)
{	
	levelWidth = theLevel.levelWidth;
	levelHeight = theLevel.levelHeight;
	//tileID = theLevel.tileID;
	
	AllocateLevelMap();
	UInt32 mapIndex;
	for (short row = 0; row < levelHeight; row++) {
		for (short column = 0; column < levelWidth; column++) {
			mapIndex = (row * levelWidth) + column;
			levelMap[mapIndex].SetValue(theLevel.levelMap[mapIndex].GetValue()) ;
		}
	}

}
	
// Destructor
GameLevel::~GameLevel(void)
{
	DeleteLevelMap();
}

// Overload assignment operator
GameLevel& GameLevel::operator= (const GameLevel& theLevel)
{
	if (this != &theLevel) { 		// Avoid self assignment
		DeleteLevelMap();
		levelWidth = theLevel.levelWidth;
		levelHeight = theLevel.levelHeight;
	
		AllocateLevelMap();
		UInt32 mapIndex;
		for (short row = 0; row < levelHeight; row++) {
			for (short column = 0; column < levelWidth; column++) {
				mapIndex = (row * levelWidth) + column;
				levelMap[mapIndex].SetValue(theLevel.levelMap[mapIndex].GetValue());
				//levelMap[row][column].SetValue(theLevel.levelMap[row][column].GetValue());
			}
		}
	}
	
	return *this;
}

// Accessor Functions
short GameLevel::GetLevelWidth(void) 
{ 
	return levelWidth;
}

void GameLevel::SetLevelWidth(short width) 
{ 
	levelWidth = width; 
}

short GameLevel::GetLevelHeight(void) 
{ 
	return levelHeight; 
}
		
void GameLevel::SetLevelHeight(short height) 
{ 
	levelHeight = height; 
}
	
void GameLevel::AllocateLevelMap(void)
{
	size_t levelMapSize = levelWidth * levelHeight * sizeof(MapElement);
	levelMap = (MapElementPtr)NewPtr(levelMapSize);
}

void GameLevel::DeleteLevelMap(void)
{
	if (levelMap != nil) {
		DisposePtr((Ptr)levelMap);
		levelMap = nil;
	}
}

void GameLevel::InitializeLevelMap (short value)
{
	UInt32 mapIndex;
	
	for (short row = 0; row < levelHeight; row++) {
		for (short column = 0; column < levelWidth; column++) {
			mapIndex = (row * levelWidth) + column;
			levelMap[mapIndex].SetValue(value);
			levelMap[mapIndex].SetVisibleToPlayer(true);
		}
	}
}

// file related functions
Boolean GameLevel::Load(CFStringRef filename, CFStringRef extension, CFStringRef subdirectory, short levelID)
{
    // You can pass the entire name of the file as the filename argument or you can
    // separate the entire name and its file extension into the
    // filename and fileExtension arguments. If you pass the entire name as the
    // filename argument, pass NULL as the fileExtension argument.
    
    // You can normally pass NULL as the subdirectory. If you place all your
    // texture files in a special directory in your application, you can pass
    // the directory name as the subdirectory argument to make searching faster.
    
    CFBundleRef mainBundle;
    CFURLRef levelURL;
    FSRef levelFile;
    
    mainBundle = CFBundleGetMainBundle();
    levelURL = CFBundleCopyResourceURL(mainBundle, filename, extension, subdirectory);
    
    if (CFURLGetFSRef(levelURL, &levelFile)) {
        return ReadLevelData(&levelFile, levelID);
    }
    else {
        return false;
    }
}

Boolean GameLevel::ReadLevelData(FSRef* filename, short levelID)
{
    // This routine has been modified so that it works on both
    // Intel and PowerPC Macs. I created the level for this level
    // back in 2001, which makes things easier because I know the
    // data is big endian. After reading every piece of data, I call the
    // function CFSwapInt16BigToHost. This routine will do nothing on
    // PowerPC Macs and swap the bytes on Intel Macs.
	
    short error;
	short refNum;
	Handle savedLevel;
	
	refNum = FSOpenResFile(filename, fsRdPerm);
    UseResFile(refNum);
    
	savedLevel = Get1Resource(kLevelResourceType, levelID);
	error = ResError();
	if ((error != noErr) || (savedLevel == nil))
		return false;
		
	// Now let's set the level data
	short version;
	short theLevelWidth;
	short theLevelHeight;
	Size offset = 0;

	BlockMoveData(*savedLevel + offset, &version, k2Bytes);
	offset = offset + k2Bytes;
    version = CFSwapInt16BigToHost(version);
	
	// If resource is obsolete, don't load it.
	if (version != kCurrentLevelVersion) {
		ReleaseResource(Handle(savedLevel));
		//CloseResFile(refNum);
		return false;
	}
	
	BlockMoveData(*savedLevel + offset, &theLevelWidth, k2Bytes);
	offset = offset + k2Bytes;
    theLevelWidth = CFSwapInt16BigToHost(theLevelWidth);
    
	BlockMoveData(*savedLevel + offset, &theLevelHeight, k2Bytes);
	offset = offset + k2Bytes;
    theLevelHeight = CFSwapInt16BigToHost(theLevelHeight);
    
	// There is currently no need for storing the tile list's resource ID
    // since the code reads tile lists from external files. But the ID is
    // still in the level data file so skip over it.
	offset = offset + k2Bytes;
	
	SetLevelWidth(theLevelWidth);
	SetLevelHeight(theLevelHeight);
	
	// Now that we have the level width and level height, allocate
	// the level map.
	AllocateLevelMap();
	
	// Read the level map
    UInt32 mapIndex;
    short mapValue;
    
	for(short row = 0; row < theLevelHeight; row++) {
		for(short column = 0; column < theLevelWidth; column++) {
			mapIndex = (row * theLevelWidth) + column;
            // Read the tile number.
            BlockMoveData(*savedLevel + offset, &mapValue, k2Bytes);
            offset = offset + k2Bytes;
            mapValue = CFSwapInt16BigToHost(mapValue);

            // Each tile number is two bytes long, with two bytes of padding.
            // Skip the padding. This is covering for a flaw in the level
            // editor I wrote 5 years ago to make the level. I wrote out the MapElement
            // data structure instead of writing 2 bytes for each tile number
            // that makes up the level.
            offset = offset + k2Bytes;

            levelMap[mapIndex].SetValue(mapValue);
            
            // My original plan for the Mac Game Programming code was
            // to make areas of the level black until the player reached
            // the area. That is why there is a visibleToPlayer data member
            // in the MapElement class, even though this code doesn't use it.
            levelMap[mapIndex].SetVisibleToPlayer(true);
		}
	}

	// Clean up
	ReleaseResource(Handle(savedLevel));
	error = ResError();
	if (SeriousError(error))
		return false;
	
	return true;
	
}


// Non-member functions
Boolean SeriousError(OSErr errorCode)
{
	switch (errorCode) {
		case noErr:
		case dupFNErr:
		case opWrErr:
			 	return false;
		default:
				return true;
				
	}
}
